#!/usr/bin/python

import os
exitcode = os.system('run-on-tsi ~/bin/run-vasp5-lock >/dev/null')
#exitcode = os.system('run-on-tsi /opt/bin/run-vasp5 >/dev/null')

#exitcode = os.system('~/bin/run-vasp5-lock')
